from .fs import DECODEX_DIR


NULL_ADDRESS_0x0 = "0x0000000000000000000000000000000000000000"
NULL_ADDRESS_0xF = "0xffffffffffffffffffffffffffffffffffffffff"

__all__ = [
    "NULL_ADDRESS_0x0",
    "NULL_ADDRESS_0xF"
    #
    "DECODEX_DIR",
]
